/*
Test file for shared ngspice
Copyright Holger Vogt 2017
New BSD license

ngspice library loaded dynamically
simple manual input
*/

/* The original function pointers

ret = ((int (*)(SendChar*, SendStat*, ControlledExit*, SendData*, SendInitData*,
BGThreadRunning*, void*)) ngSpice_Init_handle)(ng_getchar, ng_getstat,
ng_exit, NULL, ng_initdata, ng_thread_runs, NULL);

ret = ((int (*)(char*)) ngSpice_Command_handle)("source ../../examples/adder_mos.cir");

curplot = ((char * (*)()) ngSpice_CurPlot_handle)();

vecarray = ((char ** (*)(char*)) ngSpice_AllVecs_handle)(curplot);

myvec = ((pvector_info (*)(char*)) ngSpice_GVI_handle)(plotvec);

ret = ((int (*)(char**)) ngSpice_Circ_handle)(circarray);

may be replaced by their shortcuts

ret = ngSpice_Init_handle(ng_getchar, ng_getstat,
ng_exit, NULL, ng_initdata, ng_thread_runs, NULL);

ret = ngSpice_Command_handle("source ../../examples/adder_mos.cir");

curplot = ngSpice_CurPlot_handle();

vecarray = ngSpice_AllVecs_handle(curplot);

myvec = ngSpice_GVI_handle(plotvec);

ret = ngSpice_Circ_handle(circarray);
*/



#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#ifndef _MSC_VER
#include <stdbool.h>
#include <pthread.h>
#else
#define bool int
#define true 1
#define false 0
#define strdup _strdup
#endif
#include "../include/sharedspice.h"

#if defined(__MINGW32__) ||  defined(_MSC_VER)
#undef BOOLEAN
#include <windows.h>
typedef FARPROC funptr_t;
void *dlopen (const char *, int);
funptr_t dlsym (void *, const char *);
int dlclose (void *);
char *dlerror (void);
#define RTLD_LAZY	1	/* lazy function call binding */
#define RTLD_NOW	2	/* immediate function call binding */
#define RTLD_GLOBAL	4	/* symbols in this dlopen'ed obj are visible to other dlopen'ed objs */
static char errstr[128];
#else
#include <dlfcn.h> /* to load libraries*/
#include <unistd.h>
#include <ctype.h>
typedef void *  funptr_t;
#endif

bool no_bg = true;
bool not_yet = true;
bool will_unload = false;
bool error_ngspice = false;

int cieq(register char *p, register char *s);
int ciprefix(const char *p, const char *s);
int getLine(char *prmpt, char *buff, size_t sz);

/* callback functions used by ngspice */
int
ng_getchar(char* outputreturn, int ident, void* userdata);

int
ng_getstat(char* outputreturn, int ident, void* userdata);

int
ng_thread_runs(bool noruns, int ident, void* userdata);

ControlledExit ng_exit;
SendData ng_data;
SendInitData ng_initdata;

//int vecgetnumber = 0;
//double v2dat;
//static bool has_break = false;
char comd[1024];

/* functions exported by ngspice */
funptr_t ngSpice_Init_handle = NULL;
funptr_t ngSpice_Command_handle = NULL;
funptr_t ngSpice_Circ_handle = NULL;
funptr_t ngSpice_CurPlot_handle = NULL;
funptr_t ngSpice_AllVecs_handle = NULL;
funptr_t ngSpice_GVI_handle = NULL;

void * ngdllhandle = NULL;

#ifndef _MSC_VER
pthread_t mainthread;
#endif // _MSC_VER

int main()
{
    char *errmsg = NULL, *loadstring;
    int ret;

    printf("****************************\n");
    printf("**  ngspice shared start  **\n");
    printf("****************************\n");

#ifndef _MSC_VER
    mainthread = pthread_self();
#endif // _MSC_VER
    printf("Load ngspice.dll\n");
#ifdef __CYGWIN__
    loadstring = "/cygdrive/c/cygwin/usr/local/bin/cygngspice-0.dll";
#elif _MSC_VER
//    loadstring = "ngspice.dll";
//    loadstring = "..\\..\\dlls\\ngspice-26.dll";
    loadstring = "..\\Debug.Win32\\ngspice.dll";
#elif __MINGW32__
    loadstring = "D:\\Spice_general\\ngspice-master\\visualc\\sharedspice\\ReleaseOMP.x64\\ngspice.dll";
#else
    loadstring = "libngspice.so";
#endif
    ngdllhandle = dlopen(loadstring, RTLD_NOW);
    errmsg = dlerror();
    if (errmsg)
        printf("%s\n", errmsg);
    if (ngdllhandle)
        printf("ngspice.dll loaded\n");
    else {
        printf("ngspice.dll not loaded !\n");
        getLine("Press any key to exit: ", comd, sizeof(comd));
        return 1;
    }

    ngSpice_Init_handle = dlsym(ngdllhandle, "ngSpice_Init");
    errmsg = dlerror();
    if (errmsg)
        printf(errmsg);
    ngSpice_Command_handle = dlsym(ngdllhandle, "ngSpice_Command");
    errmsg = dlerror();
    if (errmsg)
        printf(errmsg);
    ngSpice_CurPlot_handle = dlsym(ngdllhandle, "ngSpice_CurPlot");
    errmsg = dlerror();
    if (errmsg)
        printf(errmsg);
    ngSpice_AllVecs_handle = dlsym(ngdllhandle, "ngSpice_AllVecs");
    errmsg = dlerror();
    if (errmsg)
        printf(errmsg);
    ngSpice_GVI_handle = dlsym(ngdllhandle, "ngGet_Vec_Info");
    errmsg = dlerror();
    if (errmsg)
        printf(errmsg);

    ret = ngSpice_Init_handle(ng_getchar, ng_getstat,
                              ng_exit, NULL, ng_initdata, ng_thread_runs, NULL);

    error_ngspice = false;
    will_unload = false;

    for (;;) {
        /* get command from stdin */
        getLine("Command: ", comd, sizeof(comd));

        /* return upon 'exit' */
        if (cieq("exit", comd))
            break;

        /* If command 'bg_run' is given, ngSpice_Command_handle() will return immediately.
           To guarantee that the primary thread here waits until the run is finished, we
           may set no_bg to 0 already here. Risk: if starting the simulation fails, we never
           may leave the waiting loop. As an alternative callback function ng_thread_runs()
           will set no_bg to 0. This has to happen within the first 200ms waiting time. */
        if (cieq("bg_run", comd))
            no_bg = false;

        ret = ngSpice_Command_handle(comd);
        if (ret) {
            /* unload now */
            dlclose(ngdllhandle);
            ngdllhandle = NULL;
            printf("ngspice unloaded\n\n");
            getLine("Press any key to exit: ", comd, sizeof(comd));
            break;
        }
        /* wait until simulation finishes */
        for (;;) {
#if defined(__MINGW32__) || defined(_MSC_VER)
            Sleep(200);
#else
            usleep(200000);
#endif
            /* after 200ms the the callback function ng_thread_runs() should have
               set no_bg to 0, otherwise we would not wait for the end of the
               background thread.*/
            if (no_bg)
                break;
        }
    }
    return 0;
}


/* Callback function called from bg thread in ngspice to transfer
   any string created by printf or puts. Output to stdout in ngspice is
   preceded by token stdout, same with stderr.*/
int
ng_getchar(char* outputreturn, int ident, void* userdata)
{
    printf("%s\n", outputreturn);
    /* setting a flag if an error message occurred */
    if (ciprefix("stderr Error:", outputreturn))
        error_ngspice = true;
    return 0;
}

/* Callback function called from bg thread in ngspice to transfer
   simulation status (type and progress in percent). */
int
ng_getstat(char* outputreturn, int ident, void* userdata)
{
    printf("%s\n", outputreturn);
    return 0;
}

/* Callback function called from ngspice upon starting (returns true) or
  leaving (returns false) the bg thread. */
int
ng_thread_runs(bool noruns, int ident, void* userdata)
{
    no_bg = noruns;
    if (noruns)
        printf("\nbg not running\n");
    else
        printf("bg running\n\n");

    return 0;
}

/* Callback function called from bg thread in ngspice if fcn controlled_exit()
   is hit. Do not exit, but unload ngspice. */
int
ng_exit(int exitstatus, bool immediate, bool quitexit, int ident, void* userdata)
{
    /*
        if(quitexit) {
            printf("DNote: Returned from quit with exit status %d\n", exitstatus);
        }
        if(immediate) {
            printf("DNote: Unload ngspice\n");
            ((int * (*)(char*)) ngSpice_Command_handle)("quit");
            dlclose(ngdllhandle);
        }

        else {
            printf("DNote: Prepare unloading ngspice\n");
            will_unload = true;
        }
    */
    return exitstatus;

}

/* Callback function called from bg thread in ngspice once per accepted data point */
int
ng_data(pvecvaluesall vdata, int numvecs, int ident, void* userdata)
{
    /*
        int *ret;

        v2dat = vdata->vecsa[vecgetnumber]->creal;
        if (!has_break && (v2dat > 0.5)) {
        // using signal SIGTERM by sending to main thread, alterp() then is run from the main thread,
        //(not on Windows though!)
    #ifndef _MSC_VER
            if (testnumber == 4)
                pthread_kill(mainthread, SIGTERM);
    #endif
            has_break = true;
        // leave bg thread for a while to allow halting it from main
    #if defined(__MINGW32__) || defined(_MSC_VER)
            Sleep (100);
    #else
            usleep (100000);
    #endif
    //        ret = ((int * (*)(char*)) ngSpice_Command_handle)("bg_halt");
        }
    */
    return 0;
}


/* Callback function called from bg thread in ngspice once upon intialization
   of the simulation vectors)*/
int
ng_initdata(pvecinfoall intdata, int ident, void* userdata)
{
    /*
        int i;
        int vn = intdata->veccount;
        for (i = 0; i < vn; i++) {
            printf("Vector: %s\n", intdata->vecs[i]->vecname);
            // find the location of V(2)
            if (cieq(intdata->vecs[i]->vecname, "V(2)"))
                vecgetnumber = i;
        }
    */
    return 0;
}

/* Unify LINUX and Windows dynamic library handling:
   Add functions dlopen, dlsym, dlerror, dlclose to Windows by
   tranlating to Windows API functions.
*/
#if defined(__MINGW32__) ||  defined(_MSC_VER)

void *dlopen(const char *name,int type)
{
    return LoadLibrary((LPCSTR)name);
}

funptr_t dlsym(void *hDll, const char *funcname)
{
    return GetProcAddress(hDll, funcname);
}

char *dlerror(void)
{
    LPVOID lpMsgBuf;
    char * testerr;
    DWORD dw = GetLastError();

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0,
        NULL
    );
    testerr = (char*)lpMsgBuf;
    strcpy(errstr,lpMsgBuf);
    LocalFree(lpMsgBuf);
    if (ciprefix("Der Vorgang wurde erfolgreich beendet.", errstr))
        return NULL;
    else
        return errstr;
}

int dlclose (void *lhandle)
{
    return (int)FreeLibrary(lhandle);
}
#endif

/* Case insensitive str eq.
   Like strcasecmp( ) XXX */
int
cieq(register char *p, register char *s)
{
    while (*p) {
        if ((isupper(*p) ? tolower(*p) : *p) !=
                (isupper(*s) ? tolower(*s) : *s))
            return(false);
        p++;
        s++;
    }
    return (*s ? false : true);
}

/* Case insensitive prefix. */
int
ciprefix(const char *p, const char *s)
{
    while (*p) {
        if ((isupper(*p) ? tolower(*p) : *p) !=
                (isupper(*s) ? tolower(*s) : *s))
            return(false);
        p++;
        s++;
    }
    return (true);
}

/* read a line from console input
   source:
   https://stackoverflow.com/questions/4023895/how-to-read-string-entered-by-user-in-c
   */
#define OK       0
#define NO_INPUT 1
#define TOO_LONG 2
int
getLine(char *prmpt, char *buff, size_t sz)
{
    int ch, extra;

    // Get line with buffer overrun protection.
    if (prmpt != NULL) {
        printf("%s", prmpt);
        fflush(stdout);
    }
    if (fgets(buff, sz, stdin) == NULL)
        return NO_INPUT;

    // If it was too long, there'll be no newline. In that case, we flush
    // to end of line so that excess doesn't affect the next call.
    if (buff[strlen(buff) - 1] != '\n') {
        extra = 0;
        while (((ch = getchar()) != '\n') && (ch != EOF))
            extra = 1;
        return (extra == 1) ? TOO_LONG : OK;
    }

    // Otherwise remove newline and give string back to caller.
    buff[strlen(buff) - 1] = '\0';
    return OK;
}
